// vektorski podatkovni tipi
__kernel void dot_product(__global const float4 *A,
						  __global const float4 *B,
						  __global float *P,
						  int size,
						  __local float4 *partial)
{
	int lid = get_local_id(0);
	int gid = get_global_id(0);

	float4 sum = (float4)(0.0f);
    while (gid < size/4)
    {
		sum += A[gid] * B[gid];
       	gid += get_global_size(0);
    }
	partial[lid] = sum;

	barrier(CLK_LOCAL_MEM_FENCE);

	int floorPow2 = get_local_size(0);
	while (floorPow2 & (floorPow2-1))
		floorPow2 &= floorPow2-1;
	if (get_local_size(0) != floorPow2)										
	{
		if (lid >= floorPow2)
            partial[lid - floorPow2] += partial[lid];
		barrier(CLK_LOCAL_MEM_FENCE);
    }

	for (int idxStep = floorPow2>>1; idxStep > 0; idxStep >>= 1)
	{
		if (lid < idxStep)
			partial[lid] += partial[lid+idxStep];
		barrier(CLK_LOCAL_MEM_FENCE);
	}

	if (lid == 0)
	{		
		P[get_group_id(0)] = dot(partial[0], (float4)(1.0f));
	}
}
