// lokalno seštevanje, skalarji
// vektorski podatkovni tipi

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <CL/cl.h>

#define MAX_SOURCE_SIZE (16384)

int main(int argc, char *argv[])
{
	char ch;
	int i;
	cl_int ret;
    FILE *fp;
    char *source_str;
    size_t source_size;

	// read params
	if (argc < 4)
	{
		printf("%s <problem size> <workgroup size> <kernel ID>\n", argv[0]);
		return 1;
	}
	int vectorSize = atoi(argv[1]);
	if (vectorSize == 0)
		vectorSize = 1024;
	int WORKGROUP_SIZE = atoi(argv[2]);
	if (WORKGROUP_SIZE == 0)
		WORKGROUP_SIZE = 256;

	// Rezervacija pomnilnika
	float *A = (float *)malloc(vectorSize * sizeof(float));
	float *B = (float *)malloc(vectorSize * sizeof(float));

	// Inicializacija vektorjev
	srand((int)time(NULL));
	for (i = 0; i < vectorSize; i++)
	{
		A[i] = rand()/(double)RAND_MAX;
		B[i] = rand()/(double)RAND_MAX;
	}

    // branje datoteke
	char fileName[100];
	sprintf(fileName, "%s-%s.cl", argv[0], argv[3]);
    fp = fopen(fileName, "r");
    if(!fp)
    {
        fprintf(stderr, ":-(\n");
        return 1;
    }
    source_str = (char*)malloc(MAX_SOURCE_SIZE);
    source_size = fread(source_str, 1, MAX_SOURCE_SIZE, fp);
    source_str[source_size] = '\0';
    fclose(fp);

	// Podatki o platformi
	cl_platform_id	platform_id[10];
	cl_uint			ret_num_platforms;
	char			*buf;
	size_t			buf_len;
	ret = clGetPlatformIDs(10, platform_id, &ret_num_platforms);
	// max. "stevilo platform, kazalec na platforme, dejansko "stevilo platform

	// Podatki o napravi
	cl_device_id	device_id[10];
	cl_uint			ret_num_devices;
	// Delali bomo s platform_id[0] na GPU
	ret = clGetDeviceIDs(platform_id[0], CL_DEVICE_TYPE_GPU, 10, device_id, &ret_num_devices);
	// izbrana platforma, tip naprave, koliko naprav nas zanima
	// kazalec na naprave, dejansko "stevilo naprav

	// Kontekst
	cl_context context = clCreateContext(NULL, 1, &device_id[0], NULL, NULL, &ret);
	// kontekst: vklju"cene platforme - NULL je privzeta, "stevilo naprav, 
	// kazalci na naprave, kazalec na call-back funkcijo v primeru napake
	// dodatni parametri funkcije, "stevilka napake

	// Ukazna vrsta
	cl_command_queue command_queue = clCreateCommandQueue(context, device_id[0], CL_QUEUE_PROFILING_ENABLE, &ret);
	// kontekst, naprava, INORDER/OUTOFORDER, napake

	// Delitev dela
	size_t local_item_size = WORKGROUP_SIZE;
	size_t num_groups = ((vectorSize/4 - 1) / local_item_size + 1);
	size_t global_item_size = num_groups * local_item_size;

	float *P = (float *)malloc(num_groups * sizeof(float));

	// Alokacija pomnilnika na napravi
	cl_mem a_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, vectorSize * sizeof(float), A, &ret);
	cl_mem b_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, vectorSize * sizeof(float), B, &ret);
	cl_mem p_mem_obj = clCreateBuffer(context, CL_MEM_WRITE_ONLY, num_groups * sizeof(float), NULL, &ret);

	// Priprava programa
	cl_program program = clCreateProgramWithSource(context, 1, (const char **)&source_str, NULL, &ret);
	// kontekst, "stevilo kazalcev na kodo, kazalci na kodo,		
	// stringi so NULL terminated, napaka													

	// Prevajanje
	ret = clBuildProgram(program, 1, &device_id[0], NULL, NULL, NULL);
	// program, "stevilo naprav, lista naprav, opcije pri prevajanju,
	// kazalec na funkcijo, uporabni"ski argumenti

	// Log
	size_t build_log_len;
	char *build_log;
	ret = clGetProgramBuildInfo(program, device_id[0], CL_PROGRAM_BUILD_LOG, 0, NULL, &build_log_len);
	// program, "naprava, tip izpisa, 
	// maksimalna dol"zina niza, kazalec na niz, dejanska dol"zina niza
	build_log = (char *)malloc(sizeof(char)*(build_log_len + 1));
	ret = clGetProgramBuildInfo(program, device_id[0], CL_PROGRAM_BUILD_LOG, build_log_len, build_log, NULL);
	printf("%s\n", build_log);
	free(build_log);
	if(build_log_len > 2)	
		return 1;

	// "s"cepec: priprava objekta
	cl_kernel kernel = clCreateKernel(program, "dot_product", &ret);
	// program, ime "s"cepca, napaka

	// "s"cepec: argumenti
	ret = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&a_mem_obj);
	ret |= clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&b_mem_obj);
	ret |= clSetKernelArg(kernel, 2, sizeof(cl_mem), (void *)&p_mem_obj);
	ret |= clSetKernelArg(kernel, 3, sizeof(cl_int), (void *)&vectorSize);
	// dinamična alokacija lokalnega pomnilnika
	ret |= clSetKernelArg(kernel, 4, local_item_size*4*sizeof(float), NULL);		
	// "s"cepec, "stevilka argumenta, velikost podatkov, kazalec na podatke

	cl_event event;
	// "s"cepec: zagon
	ret = clEnqueueNDRangeKernel(command_queue, kernel, 1, NULL, &global_item_size, &local_item_size, 0, NULL, &event);
	// vrsta, "s"cepec, dimenzionalnost, mora biti NULL, 
	// kazalec na "stevilo vseh niti, kazalec na lokalno "stevilo niti, 
	// dogodki, ki se morajo zgoditi pred klicem

	// Kopiranje rezultatov
	ret = clEnqueueReadBuffer(command_queue, p_mem_obj, CL_TRUE, 0, num_groups * sizeof(float), P, 0, NULL, NULL);
	// branje v pomnilnik iz naparave, 0 = offset
	// zadnji trije - dogodki, ki se morajo zgoditi prej


	// timing
	cl_ulong time_start;
	cl_ulong time_end;

	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start,	NULL);
	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(time_end), &time_end, NULL);

	// "ci"s"cenje
	ret = clFlush(command_queue);
	ret = clFinish(command_queue);
	ret = clReleaseKernel(kernel);
	ret = clReleaseProgram(program);
	ret = clReleaseMemObject(a_mem_obj);
	ret = clReleaseMemObject(b_mem_obj);
	ret = clReleaseMemObject(p_mem_obj);
	ret = clReleaseCommandQueue(command_queue);
	ret = clReleaseContext(context);

	// Prikaz rezultatov
	// CPU
	float dotProductCPU = 0.0;
	for (i=0; i<vectorSize; i++)
		dotProductCPU += A[i]*B[i];
	// GPU
	float dotProductGPU = 0.0;
	for (i=0; i<num_groups; i++)
		dotProductGPU += P[i];
	printf("Size: %d\n\tCPU: %f\n\tGPU: %f [%0.3lf ms]\n", vectorSize, dotProductCPU, dotProductGPU, (double)(time_end-time_start)/1000000);

	free(A);
	free(B);
	free(P);

	return 0;
}
